/*
 * Copyright (c) 1995, 1996, 1997, 1998 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 *
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "xfs_locl.h"

RCSID("$Id: xfs_vfsops-bsd.c,v 1.5 1998/07/23 06:25:46 assar Exp $");

/*
 * XFS vfs operations.
 */

#include <xfs_common.h>
#include <xfs_message.h>
#include <xfs_fs.h>
#include <xfs_dev.h>
#include <xfs_deb.h>
#include <xfs_vfsops.h>

static struct vnode *make_dead_vnode(struct mount * mp);

static int
xfs_mount(struct mount * mp,
	  char *user_path,
	  caddr_t user_data,
	  struct nameidata * ndp,
	  struct proc * p)
{
    return xfs_mount_common(mp, user_path, user_data, ndp, p);
}

static int
xfs_start(struct mount * mp, int flags, struct proc * p)
{
    XFSDEB(XDEBVFOPS, ("xfs_start mp = %p, flags = %d, proc = %p\n", 
		       mp, flags, p));
    return 0;
}


static int
xfs_unmount(struct mount * mp, int mntflags, struct proc *p)
{
    XFSDEB(XDEBVFOPS, ("xfs_umount: mp = %p, mntflags = %d, proc = %p\n", 
		       mp, mntflags, p));
    return xfs_unmount_common(mp, mntflags);
}

static int
xfs_root(struct mount * mp, struct vnode ** vpp)
{
    struct xfs *xfsp = VFS_TO_XFS(mp);
    struct xfs_message_getroot msg;
    int error;

    XFSDEB(XDEBVFOPS, ("xfs_root mp = %p\n", mp));

    do {
	if (xfsp->root != NULL) {
	    *vpp = XNODE_TO_VNODE(xfsp->root);
	    VREF(*vpp);
	    return 0;
	}
	msg.header.opcode = XFS_MSG_GETROOT;
	msg.cred.uid = curproc->p_ucred->cr_uid;
	msg.cred.pag = 0;	       /* XXX */
	error = xfs_message_rpc(xfsp->fd, &msg.header, sizeof(msg));
	if (error == 0)
	    error = ((struct xfs_message_wakeup *) & msg)->error;
    } while (error == 0);
    /*
     * Failed to get message through, need to pretend that all went well
     * and return a fake dead vnode to be able to unmount.
     */
    *vpp = make_dead_vnode(mp);
    (*vpp)->v_flag |= VROOT;
    return 0;
}


static int
xfs_quotactl(struct mount *mp, int cmd, uid_t uid, caddr_t arg, struct proc *p)
{
    XFSDEB(XDEBVFOPS, ("xfs_quotactl: mp = %p, cmd = %d, uid = %u, "
		       "arg = %p, proc = %p\n", 
		       mp, cmd, uid, arg, p));
    return EOPNOTSUPP;
}

static int
xfs_statfs(struct mount *mp, struct statfs *sbp, struct proc *p)
{
    XFSDEB(XDEBVFOPS, ("xfs_statfs: mp = %p, sbp = %p, proc = %p\n", 
		       mp, sbp, p));
    bcopy(&mp->mnt_stat, sbp, sizeof(*sbp));
    return 0;
}

static int
xfs_sync(struct mount *mp, int waitfor, struct ucred *cred, struct proc *p)
{
    XFSDEB(XDEBVFOPS, ("xfs_sync: mp = %p, waitfor = %d, "
		       "cred = %p, proc = %p\n",
		       mp, waitfor, cred, p));
    return 0;
}

static int
xfs_vget(struct mount * mp,
	 ino_t ino,
	 struct vnode ** vpp)
{
    XFSDEB(XDEBVFOPS, ("xfs_vget\n"));
    return EOPNOTSUPP;
}

static int
xfs_fhtovp(struct mount * mp,
	   struct fid * fhp,
	   struct mbuf * nam,
	   struct vnode ** vpp,
	   int *exflagsp,
	   struct ucred ** credanonp)
{
    XFSDEB(XDEBVFOPS, ("xfs_fhtovp\n"));
    return EOPNOTSUPP;
}

static int
xfs_vptofh(struct vnode * vp,
	   struct fid * fhp)
{
    XFSDEB(XDEBVFOPS, ("xfs_vptofh\n"));
    return EOPNOTSUPP;
}

#if defined(__NetBSD__) || defined(__OpenBSD__)
static void
xfs_init(void)
{
    XFSDEB(XDEBVFOPS, ("xfs_init\n"));
    return;
}

#ifdef HAVE_STRUCT_VFSOPS_VFS_OPV_DESCS
extern struct vnodeopv_desc xfs_vnodeop_opv_desc;
struct vnodeopv_desc *xfs_vnodeopv_descs[] = {
    &xfs_vnodeop_opv_desc,
    NULL,
};
#endif

static struct vfsops xfs_vfsops = {
    "xfs",
    xfs_mount,
    xfs_start,
    xfs_unmount,
    xfs_root,
    xfs_quotactl,
    xfs_statfs,
    xfs_sync,
    xfs_vget,
    xfs_fhtovp,
    xfs_vptofh,
    xfs_init,
#ifdef HAVE_STRUCT_VFSOPS_VFS_OPV_DESCS
    0,
    0,
    xfs_vnodeopv_descs,
#endif
};

/*
 * To be able to unmount when the XFS daemon is not
 * responding we need a root vnode, use a dead vnode!
 */

extern int (**dead_vnodeop_p) (void *);

/*
 *
 */
static int
xfs_uprintf_filsys(void)
{
#if 0
    {
	int i;

	uprintf("Currently loaded filesystems are:\n");
	for (i = 0; i < nvfssw; i++)
	    if (vfssw[i] != NULL)
		uprintf("vfssw[%d] = { \"%s\", %p}\n",
			i, vfssw[i]->vfs_name, vfssw[i]);
    }
#endif
#if 0
    {
	int i;
	struct vnode *t;

	for (i = 0; i < NXFS; i++) {
	    uprintf("Current xfs_nodes on xfs[%d] are:\n", i);

	    if (XFS_TO_VFS(&xfs[i]) == NULL)
		continue;

	    t = XFS_TO_VFS(&xfs[i])->mnt_vnodelist.lh_first;
	    while (t != NULL) {
		struct xfs_node *x = VNODE_TO_XNODE(t);

		uprintf("%d.%d.%d.%d(%d) ",
			x->handle.a,
			x->handle.b,
			x->handle.c,
			x->handle.d,
			x->vn->v_usecount);
		uprintf(" !\n");
		t = t->v_mntvnodes.le_next;
	    }
	}
    }
#endif
    return 0;
}

/*
 * Install and uninstall filesystem.
 */

extern struct vnodeopv_desc xfs_vnodeop_opv_desc;

#if defined(HAVE_KERNEL_VFS_ATTACH)
int
xfs_install_filesys(void)
{
    return vfs_attach(&xfs_vfsops);
}

int
xfs_uninstall_filesys(void)
{
    return vfs_detach(&xfs_vfsops);
}
#elif defined(HAVE_VFSSW)
static int vfs_offset;

int
xfs_install_filesys(void)
{
    int i;

    XFSDEB(XDEBLKM, ("Got to begining of loading xfs_filesys\n"));
    XFSDEB(XDEBLKM, ("Checking for double loading of filesystem\n"));

    /* make sure there's no VFS in the table with this name */
    for (i = 0; i < nvfssw; i++)
	if (vfssw[i] != (struct vfsops *) 0 &&
	    strncmp(vfssw[i]->vfs_name,
		    xfs_vfsops.vfs_name,
		    MFSNAMELEN) == 0)
	    return (EEXIST);

    XFSDEB(XDEBLKM, ("Searching the last available slot\n"));

    /* pick the last available empty slot */
    for (i = nvfssw - 1; i >= 0; i--)
	if (vfssw[i] == (struct vfsops *) 0)
	    break;
    if (i < 0) {		       /* or if none, punt */
	return (EINVAL);
    }
    /*
     * Set up file system
     */
    XFSDEB(XDEBLKM, ("Setting up filesystem\n"));

    vfssw[i] = &xfs_vfsops;
    vfssw[i]->vfs_refcount = 0;

    /*
     * Call init function for this VFS...
     */
    XFSDEB(XDEBLKM, ("Doing vfs_init()\n"));
    (*(vfssw[i]->vfs_init)) ();

    /* done! */
    vfs_offset = i;

    XFSDEB(XDEBLKM, ("Using VFS slot %d\n", i));

    return 0;
}

int
xfs_uninstall_filesys(void)
{
    /* replace current slot contents with old contents */
    vfssw[vfs_offset] = (struct vfsops *) 0;

    return 0;
}
#else /* !HAVE_KERNEL_VFS_ATTACH && !HAVE_VFSSW */
int
xfs_install_filesys(void)
{
    struct vfsconf *vfsp;
    struct vfsconf **vfspp;


    /* Check if filesystem already known */
    for (vfspp = &vfsconf, vfsp = vfsconf;
	 vfsp;
	 vfspp = &vfsp->vfc_next, vfsp = vfsp->vfc_next)
	if (strncmp(vfsp->vfc_name,
		    "xfs", MFSNAMELEN) == 0)
	    return (EEXIST);

    /* Allocate and initialize */
    MALLOC(vfsp, struct vfsconf *, sizeof(struct vfsconf),
	   M_VFS, M_WAITOK);

    vfsp->vfc_vfsops = &xfs_vfsops;
    strncpy(vfsp->vfc_name, "xfs", MFSNAMELEN);
    vfsp->vfc_typenum = 0;
    vfsp->vfc_refcount = 0;
    vfsp->vfc_flags = 0;
    vfsp->vfc_mountroot = 0;
    vfsp->vfc_next = NULL;

    maxvfsconf++;

    /* Add to the end of the list */
    *vfspp = vfsp;

    /* Call vfs_init() */
    printf("Calling vfs_init()\n");
    (*(vfsp->vfc_vfsops->vfs_init)) (vfsp);

    /* done! */

    return 0;
}

int
xfs_uninstall_filesys(void)
{
    struct vfsconf *vfsp;
    struct vfsconf **vfspp;


    /* Find our vfsconf struct */
    for (vfspp = &vfsconf, vfsp = vfsconf;
	 vfsp;
	 vfspp = &vfsp->vfc_next, vfsp = vfsp->vfc_next)
	if (strncmp(vfsp->vfc_name,
		    "xfs",
		    MFSNAMELEN) == 0)
	    break;

    if (!vfsp)			       /* Not found */
	return (EEXIST);

    if (vfsp->vfc_refcount)	       /* In use */
	return (EBUSY);

    /* Remove from list and free  */
    *vfspp = vfsp->vfc_next;
    FREE(vfsp, M_VFS);

    maxvfsconf--;

    return 0;
}
#endif /* HAVE_VFSSW */

int
xfs_stat_filesys(void)
{
    return xfs_uprintf_filsys();
}

#elif defined(__FreeBSD__)
static int
xfs_init(void)
{
    XFSDEB(XDEBVFOPS, ("xfs_init\n"));
    return 0;
}

static struct vfsops xfs_vfsops = {
    xfs_mount,
    xfs_start,
    xfs_unmount,
    xfs_root,
    xfs_quotactl,
    xfs_statfs,
    xfs_sync,
    xfs_vget,
    xfs_fhtovp,
    xfs_vptofh,
    xfs_init
};

static struct vfsconf xfs_vfc = {
    &xfs_vfsops,
    "xfs",
    0,
    0,
    0
};

/*
 * To be able to unmount when the XFS daemon is not
 * responding we need a root vnode, use a dead vnode!
 */

extern vop_t **dead_vnodeop_p;

/*
 *
 */
static int
xfs_uprintf_filsys(void)
{
#if 1
    {
	int i;

	uprintf("Currently loaded filesystems are:\n");
	for (i = 0; i < MOUNT_MAXTYPE; i++)
	    if (vfsconf[i]->vfc_name)
		uprintf("vfsconf[%d] = { \"%s\", %p}\n",
			i, vfsconf[i]->vfc_name, vfssw[i]);
    }
#endif
#if 1
    {
	int i;

	struct vnode *t;

	for (i = 0; i < NXFS; i++) {
	    uprintf("Current xfs_nodes on xfs[%d] are:\n", i);

	    if (XFS_TO_VFS(&xfs[i]) == NULL)
		continue;

	    t = XFS_TO_VFS(&xfs[i])->mnt_vnodelist.lh_first;
	    while (t != NULL) {
		struct xfs_node *x = VNODE_TO_XNODE(t);

		uprintf("%d.%d.%d.%d(%d) ",
			x->handle.a,
			x->handle.b,
			x->handle.c,
			x->handle.d,
			x->vn->v_usecount);
		uprintf(" !\n");
		t = t->v_mntvnodes.le_next;
	    }
	}
    }
#endif
    return 0;
}

/*
 * Install and uninstall filesystem.
 */

extern struct vnodeopv_desc xfs_vnodeop_opv_desc;

int
xfs_install_filesys(void)
{
    int i;
    struct vnodeopv_desc *foo[] = {&xfs_vnodeop_opv_desc, NULL};

    for (i = 0; i < MOUNT_MAXTYPE; i++)
	if (strcmp(vfsconf[i]->vfc_name, xfs_vfc.vfc_name) == 0)
	    return EEXIST;

    for (i = MOUNT_MAXTYPE - 1; i >= 0; --i)
	if (vfsconf[i] == &void_vfsconf)
	    break;
    if (i < 0) {
	uprintf("failed to find free VFS slot\n");
	return EINVAL;
    }
    printf("Using VFS slot %d\n", i);

    xfs_vfc.vfc_index = i;
    vfsconf[i] = &xfs_vfc;
    vfssw[i] = xfs_vfc.vfc_vfsops;

    *(foo[0]->opv_desc_vector_p) = NULL;
    vfs_opv_init(foo);
    (*(vfssw[xfs_vfc.vfc_index]->vfs_init)) ();
    return 0;
}

int
xfs_uninstall_filesys(void)
{
    int i;

    /* Check for open, mounted and active vnodes */
    for (i = 0; i < NXFS; i++)
	if (xfs[i].nnodes)
	    printf("Warning (error really): There are active vnodes!\n");
    i = xfs_vfc.vfc_index;

    if (vfsconf[i]->vfc_refcount)
	printf("Warning: refcount != 0\n");

    vfsconf[i] = &void_vfsconf;
    vfssw[i] = NULL;
    return 0;
}

int
xfs_stat_filesys(void)
{
    return xfs_uprintf_filsys();
}

#endif

static struct vnode *
make_dead_vnode(struct mount *mp)
{
    struct vnode *dead;
    int error;

    XFSDEB(XDEBNODE, ("make_dead_vnode mp = %p\n", mp));

    error = getnewvnode(VT_NON, mp, dead_vnodeop_p, &dead);
    if (error) {
	panic("make_dead_vnode: getnewvnode failed: error = %d\n", error);
    }
    return dead;
}

int
xfs_may_uninstall_filesys(void)
{
    /* maybe brutal unloading would be better? */

#ifdef HAVE_STRUCT_VFSCONF_VFS_REFCOUNT
    if (vfssw[vfs_offset)->vfs_refcount != 0)
	return EBUSY;
#endif
    return 0;
}
